@@ -1,5 +1,5 @@
 /*
-  Copyright 2011 David Robillard <http://drobilla.net>
+  Copyright 2011-2015 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -14,17 +14,27 @@
   OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
 
+#include "dylib.h"
+#include "suil_config.h"
+#include "suil_internal.h"
+
 #include <QX11EmbedContainer>
+
 #undef signals
 
 #include <gtk/gtk.h>
 #include <gdk/gdkx.h>
 
-#include "./suil_config.h"
-#include "./suil_internal.h"
-
 extern "C" {
 
+typedef struct _SuilGtk2InQt4Wrapper SuilGtk2InQt4Wrapper;
+
+struct _SuilGtk2InQt4Wrapper {
+	QX11EmbedContainer* host_widget;
+	QWidget*            parent;
+	GtkWidget*          plug;
+};
+
 static void
 on_size_request(GtkWidget*      widget,
                 GtkRequisition* requisition,
@@ -43,13 +53,31 @@ on_size_allocate(GtkWidget*    widget,
 	wrap->resize(allocation->width, allocation->height);
 }
 
+static void
+wrapper_free(SuilWrapper* wrapper)
+{
+	SuilGtk2InQt4Wrapper* impl = (SuilGtk2InQt4Wrapper*)wrapper->impl;
+
+	if (impl->plug) {
+		gtk_widget_destroy(impl->plug);
+	}
+
+	if (impl->host_widget) {
+		delete impl->host_widget;
+	}
+
+	free(impl);
+}
+
 static int
 wrapper_wrap(SuilWrapper*  wrapper,
              SuilInstance* instance)
 {
-	QX11EmbedContainer* const wrap   = new QX11EmbedContainer();
-	GtkWidget* const          plug   = gtk_plug_new(wrap->winId());
-	GtkWidget* const          widget = (GtkWidget*)instance->ui_widget;
+	SuilGtk2InQt4Wrapper* const impl   = (SuilGtk2InQt4Wrapper*)wrapper->impl;
+	QWidget*                    root   = static_cast<QWidget*>(impl->parent);
+	QX11EmbedContainer* const   wrap   = new QX11EmbedContainer(root);
+	GtkWidget* const            plug   = gtk_plug_new(wrap->winId());
+	GtkWidget* const            widget = (GtkWidget*)instance->ui_widget;
 
 	gtk_container_add(GTK_CONTAINER(plug), widget);
 	gtk_widget_show_all(plug);
@@ -68,17 +96,20 @@ wrapper_wrap(SuilWrapper*  wrapper,
 	g_signal_connect(
 		G_OBJECT(plug), "size-allocate", G_CALLBACK(on_size_allocate), wrap);
 
+	impl->host_widget     = wrap;
+	impl->plug            = plug;
 	instance->host_widget = wrap;
 
 	return 0;
 }
 
-SUIL_API
+SUIL_LIB_EXPORT
 SuilWrapper*
 suil_wrapper_new(SuilHost*      host,
                  const char*    host_type_uri,
                  const char*    ui_type_uri,
-                 LV2_Feature*** features)
+                 LV2_Feature*** features,
+                 unsigned       n_features)
 {
 	/* We have to open libgtk here, so Gtk type symbols are present and will be
 	   found by the introspection stuff.  This is required at least to make
@@ -86,20 +117,31 @@ suil_wrapper_new(SuilHost*      host,
 	   type" errors.
 	*/
 	if (!host->gtk_lib) {
-		dlerror();
-		host->gtk_lib = dlopen(SUIL_GTK2_LIB_NAME, RTLD_LAZY|RTLD_GLOBAL);
+		dylib_error();
+		host->gtk_lib = dylib_open(SUIL_GTK2_LIB_NAME, DYLIB_LAZY|DYLIB_GLOBAL);
 		if (!host->gtk_lib) {
-			fprintf(stderr, "Failed to open %s (%s)\n",
-			        SUIL_GTK2_LIB_NAME, dlerror());
+			SUIL_ERRORF("Failed to open %s (%s)\n",
+			            SUIL_GTK2_LIB_NAME, dylib_error());
 			return NULL;
 		}
 		gtk_init(NULL, NULL);
 	}
 
-	SuilWrapper* wrapper = (SuilWrapper*)malloc(sizeof(SuilWrapper));
+	/* Create wrapper implementation. */
+	SuilGtk2InQt4Wrapper* const impl = (SuilGtk2InQt4Wrapper*)
+		calloc(1, sizeof(SuilGtk2InQt4Wrapper));
+
+	/* Set parent widget if given. */
+	for (unsigned i = 0; i < n_features; ++i) {
+		if (!strcmp((*features)[i]->URI, LV2_UI__parent)) {
+			impl->parent = static_cast<QWidget*>((*features)[i]->data);
+		}
+	}
+
+	SuilWrapper* wrapper = (SuilWrapper*)calloc(1, sizeof(SuilWrapper));
 	wrapper->wrap = wrapper_wrap;
-	wrapper->free = NULL;
-	wrapper->impl = NULL;
+	wrapper->free = wrapper_free;
+	wrapper->impl = impl;
 
 	return wrapper;
 }
